//+------------------------------------------------------------------+
//|                                                    gEA_test2.mq5 |
//|                           Copyright 2010, Krzysztof Hryniewiecki |
//|                                          khryniewiecki@gmail.com |
//+------------------------------------------------------------------+
#property copyright "Copyright 2010, Krzysztof Hryniewiecki"
#property link      "khryniewiecki@gmail.com"
#property version   "1.00"


#define BUILDING_INFO_STEP 16
#define BUILDING_INFO_LIMIT 256

#define DATAFILE_HEADER "gLibrary (C) 2010 Krzysztof Hryniewiecki\r\n\x001A"
#define DATAFILE_VERSION 1

#define ERR_DATAFILE_HEADER 1
#define ERR_DATAFILE_VERSION 2
#define ERR_OBJECT_LOADING 2

struct quote {
  datetime tm;
  double price;
};

class TObject;
class TCollection;
struct TObjectNotifyRec {
  TObject *ptr;
  uint counter;
};



//*******************************************************************
// Routines
//*******************************************************************
void FreeAndNil(TObject*& obj) {
  TObject *ptr;
  
  ptr = obj;
  obj = NULL;
  ptr.Free();
}

void WriteString(int file_handle, const string str) {
  FileWriteInteger(file_handle, StringLen(str), INT_VALUE);
  FileWriteString(file_handle, str);
}
string ReadString(int file_handle) {
  int c;
  
  c = FileReadInteger(file_handle, INT_VALUE);
  return FileReadString(file_handle, c);
}
void WriteStringArray(int file_handle, const string &arr[], int start = 0, int count = WHOLE_ARRAY) {
  int i;  
  if (count == WHOLE_ARRAY) count = ArraySize(arr);
  
  for (i = 0; i < count; i++)
    WriteString(file_handle, arr[start + i]);
}
void ReadStringArray(int file_handle, string &arr[], int start = 0, int count = WHOLE_ARRAY) {
  int i;
  if (count == WHOLE_ARRAY) count = ArraySize(arr);
  
  for (i = 0; i < count; i++)
    arr[start+i] = ReadString(file_handle);
}


//*******************************************************************
// TObject
//*******************************************************************
class TObject {
private:
  TObjectNotifyRec FNotify[];
  bool FLoading;
protected:
  virtual void Notify(string Operation);
  void Changed();
  virtual bool CheckDataFileVersion(long fileVersion) const;
  virtual bool DoSave(int handle);
  virtual bool DoLoad(int handle);
public:
  TObject();
  ~TObject();
  bool Loading() const;
  void Free();
  virtual void Notification(TObject *Sender, string Operation);
  void AddNotification(TObject *AObject);
  void RemoveNotification(TObject *AObject);
  void FreeNotification(TObject *AObject);
  void RemoveFreeNotification(TObject *AObject);
  bool Save(int handle);
  bool Load(int handle);
  bool Save(string fileName);
  bool Load(string fileName);
};

bool TObject::CheckDataFileVersion(long fileVersion) const { return (fileVersion == ERR_DATAFILE_VERSION); }
bool TObject::DoSave(int handle) { return true; } 
bool TObject::DoLoad(int handle) { return true; } 
void TObject::FreeNotification(TObject *AObject) { AddNotification(AObject); AObject.AddNotification(GetPointer(this)); }
void TObject::RemoveFreeNotification(TObject *AObject) { RemoveNotification(AObject); AObject.RemoveNotification(GetPointer(this)); }
void TObject::Changed() { Notify("Changed"); }
bool TObject::Loading() const { return FLoading; }
bool TObject::Save(int handle) { return DoSave(handle); } 


TObject::TObject() { 
  FLoading = false; 
}

TObject::~TObject() {
  TObject *ptr;
  int c;
  
  while ((c = ArraySize(FNotify)) > 0) {
    ptr = FNotify[c-1].ptr;
    ArrayResize(FNotify, c-1);
    if (CheckPointer(ptr) != POINTER_INVALID)
      ptr.Notification(GetPointer(this), "TObject::~TObject");
  }
}

void TObject::Notify(string Operation) {
  int i;
  
  for (i = ArraySize(FNotify)-1; i >= 0; i--) 
    if ((i < ArraySize(FNotify)) && (CheckPointer(FNotify[i].ptr) != POINTER_INVALID))
      FNotify[i].ptr.Notification(GetPointer(this), Operation);
}

void TObject::Free() {
  if (CheckPointer(GetPointer(this)) == POINTER_DYNAMIC) 
    delete GetPointer(this);
}

void TObject::Notification(TObject *Sender, string Operation) {
  int i, j, c;
  
  if (Operation == "TObject::~TObject") {
    c = ArraySize(FNotify);
    i = c-1;
    while ((i >= 0) && (FNotify[i].ptr != Sender)) i--;
    if (i >= 0) {
      for (j = i+1; j < c; j++)
        FNotify[j-1] = FNotify[j];
      ArrayResize(FNotify, c-1);    
    }
  }
}

void TObject::AddNotification(TObject *AObject) {
  int c, i;
  
  c = ArraySize(FNotify);
  i = c-1;
  while ((i >= 0) && (FNotify[i].ptr != AObject)) i--;
  if (i >= 0) FNotify[i].counter++;
  else {
    ArrayResize(FNotify, c+1);
    FNotify[c].ptr = AObject;
    FNotify[c].counter = 1;
  }
}

void TObject::RemoveNotification(TObject *AObject) {
  int c, i, j;
  
  c = ArraySize(FNotify);
  i = c-1;
  while ((i >= 0) && (FNotify[i].ptr != AObject)) i--;
  if (i < 0) return;
  FNotify[i].counter--;
  if (FNotify[i].counter == 0) {
    for (j = i+1; j < c; j++)
      FNotify[j-1] = FNotify[j];
    ArrayResize(FNotify, c-1);
  }
}


bool TObject::Load(int handle) {
  bool result;
  
  if (FLoading) {
    SetUserError(ERR_OBJECT_LOADING);
    return false;
  }
  
  FLoading = true;
  result = DoLoad(handle);
  FLoading = false;
  
  return result;
}

bool TObject::Save(string fileName) {
  int handler;
  bool result;
  
  handler = FileOpen(fileName, FILE_WRITE | FILE_BIN);
  if (handler == INVALID_HANDLE) return false;
  
  WriteString(handler, DATAFILE_HEADER);
  FileWriteLong(handler, DATAFILE_VERSION);
  
  result = Save(handler);
  
  FileClose(handler);
  return result;
}

bool TObject::Load(string fileName) {
  int handler;
  bool result;
  
  handler = FileOpen(fileName, FILE_READ | FILE_BIN);
  if (handler == INVALID_HANDLE) return false;

  if (ReadString(handler) != DATAFILE_HEADER) {
    SetUserError(ERR_DATAFILE_HEADER);
    FileClose(handler);
    return false;
  }
  
  if (!CheckDataFileVersion(FileReadLong(handler))) {
    SetUserError(ERR_DATAFILE_VERSION);
    FileClose(handler);
    return false;
  }
  
  result = Load(handler);

  FileClose(handler);
  return result;
}


//*******************************************************************
// TList
//*******************************************************************

class TList : public TObject {
private:
  TObject *FList[];
  bool FOwnsObjects;
  bool FSorted;
  bool FUnique;
protected:
  virtual int Compare(TObject *ptr1, TObject *ptr2) const;
  void QuickSort(int beg,int end);
  void BubbleSort();
public:
  TList();
  ~TList();
  virtual void Notification(TObject *Sender, string Operation);

  bool OwnsObjects() const;
  void OwnsObjects(bool AOwnsObjects);
  bool Unique() const;
  void Unique(bool AUnique);
  bool Sorted() const;
  void Sorted(bool ASorted);
  TObject* Item(int Index) const;
  void Item(int Index, TObject *AObject);
  virtual void Sort();
  int IndexOf(TObject *ptr) const;
  
  int Count() const;
  int Add(TObject *ptr);
  virtual int Insert(int Index, TObject *ptr);
  virtual TObject *Extract(int Index);
  TObject *Extract(TObject *ptr);
  void Delete(int Index);
  void Delete(TObject *ptr);
  void Remove(int Index);
  void Remove(TObject *ptr);
  void Clear();
  
  void Exchange(int index1, int index2);
  void Move(int index1, int index2);
};


int TList::Compare(TObject *ptr1, TObject *ptr2) const { return 0; }
bool TList::OwnsObjects() const { return FOwnsObjects; }
void TList::OwnsObjects(bool AOwnsObjects) { FOwnsObjects = AOwnsObjects; }
bool TList::Unique() const { return FUnique; }
bool TList::Sorted() const { return FSorted; }
TObject* TList::Item(int Index) const { return FList[Index]; }
int TList::Count() const { return ArraySize(FList); }
TObject *TList::Extract(TObject *ptr) { int i = IndexOf(ptr); if (i >= 0) return Extract(i); return NULL; }
void TList::Delete(int Index) { TObject *ptr = Extract(Index); ptr.Free(); }
void TList::Delete(TObject *ptr) { ptr = Extract(ptr); ptr.Free(); }
void TList::Remove(int Index) { if (FOwnsObjects) Delete(Index); else Extract(Index); }
void TList::Remove(TObject *ptr)  { if (FOwnsObjects) Delete(ptr); else Extract(ptr);  }


TList::TList() { 
  FOwnsObjects = false; 
  FSorted = false; 
  FUnique = false; 
}

TList::~TList() {
  int i;
  for (i = Count()-1; i >= 0; i--) 
    if ((i < Count()-1) && (CheckPointer(Item(i)) != POINTER_INVALID))
      Item(i).Notification(GetPointer(this), "TObject::~TObject");
  Clear();
}

void TList::Notification(TObject *Sender, string Operation) {
  int i, j, c;
  
  if (Operation == "TObject::~TObject")
    Extract(Sender);
  else if (FSorted && (Operation == "Changed")) {
    i = IndexOf(Sender);
    if (i >= 0) {
      c = Count();
      j = i+1;
      while ((j < c) && (Compare(FList[j], Sender) < 0)) j++;
      if (j != i+1) Move(i, j-1);
      else {
        j = i-1;
        while ((j >= 0) && (Compare(FList[j], Sender) > 0)) j--;
        if (j != i-1)
          Move(i, j+1);
      }
    }
  }
  TObject::Notification(Sender, Operation);
}

int TList::Insert(int Index, TObject *ptr) {
  int result, i;
  
  FUnique = FUnique && (IndexOf(ptr) < 0);
    
  ArrayResize(FList, Count()+1); 
  for (i = Count()-2; i >= Index; i--) 
    FList[i+1] = FList[i];
  FList[Index] = ptr;
  
  FSorted = FSorted &&
           ((Index == 0) || (Compare(FList[Index-1], FList[Index]) <= 0)) &&
           ((Index == Count()-1) || (Compare(FList[Index], FList[Index+1]) <= 0));
              
              
  ptr.AddNotification(GetPointer(this));
  ptr.Notification(GetPointer(this), "TList::Insert");
  Changed();
 
  return Index;
}

TObject* TList::Extract(int Index) {
  TObject *ptr;
  int i, c;
  
  c = Count();
  ptr = FList[Index];
  
  if (CheckPointer(ptr) != POINTER_INVALID) {
    ptr.Notification(GetPointer(this), "TList::Extract");
    ptr.RemoveNotification(GetPointer(this));
  }
  
  for (i = Index; i < c-1; i++) 
    FList[i] = FList[i+1];
  ArrayResize(FList, c-1);
  
  Changed();

  return ptr;
}

void TList::Unique(bool AUnique) {
  int i, j;
  
  if (FUnique == AUnique) return;
  if (AUnique) 
    for (i = 0; i < Count(); i++) {
      j = i+1;
      while (j < Count())
        if (FList[i] == FList[j])
          Extract(j);
        else
          j++;
    }
  FUnique = AUnique;
}

void TList::Sorted(bool ASorted) {
  if (FSorted == ASorted) return;
  if (ASorted)
    Sort();
  else 
    FSorted = false;
  Changed();
}

void TList::QuickSort(int beg,int end) {
   int    i,j,m;
   TObject *buffer;

   if((beg < 0) || (end < 0)) return;

   i = beg;
   j = end;
   while(i < end) {
      m = (beg+end) >> 1;
      while(i < j) {
         while(Compare(FList[i], FList[m]) < 0) {
           if(i == Count()-1) break;
           i++;
         }
         while(Compare(FList[j], FList[m]) > 0) {
           if(j == 0) break;
           j--;
         }
         if(i <= j) {
           buffer = FList[i];
           FList[i++] = FList[j];
           FList[j] = buffer;
               
           if(j == 0) 
             break;
           else  
             j--;
         }
      }
      
      if(beg < j) QuickSort(beg,j);
      beg = i;
      i = beg;
      j = end;
   }
}

void TList::BubbleSort() {
  int i, j, c;
  TObject *buffer;
  
  c = Count();
  for (i = 0; i < c; i++) {
    j = i+1;
    while ((j < c) ? (Compare(FList[j-1], FList[j]) < 0) : false) {
      buffer = FList[j-1];
      FList[j-1] = FList[j];
      FList[j] = buffer;
      j++;
    }
  }
}

void TList::Sort() {
  QuickSort(0, Count()-1);
  Sorted(true);
}

void TList::Item(int Index, TObject *AObject) {
  if (FList[Index] == AObject) return;
  FUnique = FUnique && (IndexOf(AObject) < 0);
  Extract(Index);
  Insert(Index, AObject);
}

int TList::IndexOf(TObject *ptr) const {
  int i, left, right, mid, c;
  
  c = Count();
  if (FSorted) {
    left = 0;
    right = c-1;
    while (left + 1 < right)
      if (Compare(ptr, FList[(left + right) >> 1]) <= 0)
        right = (left + right) >> 1;
      else
        left = (left + right) >> 1;
    while ((left <= c) && (Compare(ptr, FList[left]) <= 0)) {
      if (FList[left] == ptr)
        return left;
      left++;
    }
  } else 
    for (i = c-1; i >= 0; i--)
      if (FList[i] == ptr) 
        return i;
  
  return -1;
}
  
int TList::Add(TObject *ptr) { 
  int c;
  c = Count();
  
  if (FSorted) {
    int left, right, i;
    left = 0;
    right = c-1;

    while (left + 1 < right)
      if (Compare(ptr, FList[(left + right) >> 1]) < 0)
        right = (left + right) >> 1;
      else
        left = (left + right) >> 1;
    while ((left < c) && (Compare(FList[left], ptr) <= 0)) 
      left++;
      
    Insert(left, ptr);
    return left;
  } 

  Insert(c, ptr);
  return (c); 
}

void TList::Clear() {
  while (Count() > 0)
    Remove(Count()-1);
}
  
void TList::Exchange(int index1, int index2) {
  TObject *buffer = FList[index1];
  if (index1 == index2) return;
  
  FSorted = FSorted && (Compare(FList[index1], FList[index2]) == 0);
  FList[index1] = FList[index2];
  FList[index2] = buffer;
  Changed();
}

void TList::Move(int index1, int index2) {
  int i;
  TObject *buffer = FList[index1];
  if (index1 == index2) return;
  
  FSorted = FSorted && (Compare(FList[index1], FList[index2]) == 0);
  
  if (index1 < index2) 
    for (i = index1; i < index2; i++)
      FList[i] = FList[i+1];
  else 
    for (i = index1; i > index2; i--)
      FList[i] = FList[i-1];
  FList[index2] = buffer;
  
  Changed();
}



//*******************************************************************
// TNameableList
//*******************************************************************
class TNameableList: public TList {
protected:
  virtual int Compare(TObject *ptr1, TObject *ptr2) const;
public:
  TNameableList();
  int IndexOf(string AName) const;
  virtual string NameOf(TObject *ptr) const; // abstract
  string NameOf(int Index) const;
  TObject *Find(string AName) const;
};

string TNameableList::NameOf(TObject *ptr) const { return ""; } // abstract
string TNameableList::NameOf(int Index) const { return NameOf(Item(Index)); }

TNameableList::TNameableList() { 
  Sorted(true); 
}

TObject *TNameableList::Find(string AName) const { 
  int index = IndexOf(AName); 
  if (index < 0) return NULL; 
  return Item(index); 
}

int TNameableList::Compare(TObject *ptr1, TObject *ptr2) const {
  string name1, name2;
  
  name1 = NameOf(ptr1);
  name2 = NameOf(ptr2);
  
  if (name1 < name2) return -1;
  if (name1 > name2) return 1;
  return 0;
}

int TNameableList::IndexOf(string AName) const {
  int left, right;
  
  left = 0;
  right = Count()-1;
  while (left+1 < right)
    if (NameOf((left+right)/2) < AName) 
      left = (left+right)/2;
    else
      right = (left+right)/2;
      
  while ((left <= right) ? (NameOf(left) < AName) : false)
    left++;
    
  if ((left <= right) ? (NameOf(left) == AName) : false)
    return left;

  return -1;
}



//*******************************************************************
// TCollection and TCollectionItem
//*******************************************************************

class TCollectionItem: TObject {
private:
  TCollection *FCollection;
public:
  TCollectionItem() { FCollection = NULL; }
  TCollection *Collection() const { return FCollection; }
  virtual void Notification(TObject *Sender, string Operation);
};

class TCollection: TList {
protected:
  virtual bool DoSave(int handle);
  virtual bool DoLoad(int handle);
public:
  TCollection() { Unique(true); OwnsObjects(true); }
  virtual TCollectionItem* NewItem() const;
  virtual int Insert(int Index, TObject *ptr);
  virtual TObject *Extract(int Index);
};

TCollectionItem* TCollection::NewItem() const { return NULL; } // abstract

void TCollectionItem::Notification(TObject *Sender, string Operation) {
  if (Operation == "TCollection::Insert") {
    if ((CheckPointer(FCollection) != POINTER_INVALID) && (FCollection != Sender))
      FCollection.Extract(GetPointer(this));
    FCollection = (TCollection*)Sender;
  } else
  if ((Operation == "TCollection::Extract") && (FCollection == Sender)) 
    FCollection = NULL;
  TObject::Notification(Sender, Operation);
}


int TCollection::Insert(int Index, TObject *ptr) {
  int i;
  
  i = IndexOf(ptr);
  if (i >= 0) return i;
  
  ptr.Notification(GetPointer(this), "TCollection::Insert");
  return TList::Insert(Index, ptr);
}

TObject *TCollection::Extract(int Index) {
  Item(Index).Notification(GetPointer(this), "TCollection::Extract");
  return TList::Extract(Index);
}

bool TCollection::DoSave(int handle) { 
  int i, c = Count();
  
  if (!TList::DoSave(handle)) return false;
  
  if (FileWriteInteger(handle, (Sorted() ? 1 : 0), CHAR_VALUE) != CHAR_VALUE) return false;
  if (FileWriteInteger(handle, c, INT_VALUE) != INT_VALUE) return false;
  for (i = 0; i < c; i++)
    if (!Item(i).Save(handle)) return false;
  
  return true; 
} 

bool TCollection::DoLoad(int handle) { 
  int c, i;
  TCollectionItem *item;
  
  Clear();
  if (!TList::DoLoad(handle)) return false;
  
  Sorted(FileReadInteger(handle, CHAR_VALUE) != 0);
  c = FileReadInteger(handle, INT_VALUE);
  for (i = 0; i < c; i++) {
    item = NewItem();
    if (!item.Load(handle)) {
      item.Free();
      return false;
    }
    Add(item);
  }
  
  return true; 
} 




int OnInit() {
  // trends.Trend(Symbol());
  return(0);
}

void OnDeinit(const int reason) {
}

void OnTick() {
}

